@@ -14,6 +14,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import annotations
+
 import abc
 import enum
 import itertools
@@ -26,7 +28,8 @@ import chess
 import chess.engine
 import chess.svg
 
-from typing import Callable, Dict, Generic, Iterable, Iterator, List, Mapping, MutableMapping, Set, TextIO, Tuple, Type, TypeVar, Optional, Union
+from typing import Any, Callable, Dict, Generic, Iterable, Iterator, List, Mapping, MutableMapping, Set, TextIO, Tuple, Type, TypeVar, Optional, Union
+from chess import Square
 
 
 LOGGER = logging.getLogger(__name__)
@@ -139,7 +142,7 @@ ResultT = TypeVar("ResultT", covariant=True)
 
 
 class _AcceptFrame:
-    def __init__(self, node: "GameNode", *, is_variation: bool = False, sidelines: bool = True):
+    def __init__(self, node: GameNode, *, is_variation: bool = False, sidelines: bool = True):
         assert node.parent is not None, "cannot visit dangling GameNode"
         self.state = "pre"
         self.node = node
@@ -161,7 +164,7 @@ class GameNode:
         self.board_cached: Optional[weakref.ref[chess.Board]] = None
 
     @classmethod
-    def dangling_node(cls) -> "GameNode":
+    def dangling_node(cls) -> GameNode:
         return GameNode()
 
     def _board(self) -> chess.Board:
@@ -227,7 +230,7 @@ class GameNode:
         assert self.parent is not None and self.move is not None, "cannot get uci of dangling GameNode"
         return self.parent.board().uci(self.move, chess960=chess960)
 
-    def root(self) -> "GameNode":
+    def root(self) -> GameNode:
         node = self
         while node.parent:
             node = node.parent
@@ -239,7 +242,7 @@ class GameNode:
         assert isinstance(root, Game), "GameNode not rooted in Game"
         return root
 
-    def end(self) -> "GameNode":
+    def end(self) -> GameNode:
         """Follows the main variation to the end and returns the last node."""
         node = self
 
@@ -290,7 +293,7 @@ class GameNode:
 
         return not self.parent.variations or self.parent.variations[0] == self
 
-    def __getitem__(self, move: Union[int, chess.Move]) -> "GameNode":
+    def __getitem__(self, move: Union[int, chess.Move, GameNode]) -> GameNode:
         try:
             return self.variations[move]  # type: ignore
         except TypeError:
@@ -300,41 +303,49 @@ class GameNode:
 
         raise KeyError(move)
 
-    def variation(self, move: Union[int, chess.Move]) -> "GameNode":
+    def __contains__(self, move: Union[int, chess.Move, GameNode]) -> bool:
+        try:
+            self[move]
+        except KeyError:
+            return False
+        else:
+            return True
+
+    def variation(self, move: Union[int, chess.Move, GameNode]) -> GameNode:
         """
         Gets a child node by either the move or the variation index.
         """
         return self[move]
 
-    def has_variation(self, move: chess.Move) -> bool:
-        """Checks if the given *move* appears as a variation."""
-        return move in (variation.move for variation in self.variations)
+    def has_variation(self, move: Union[int, chess.Move, GameNode]) -> bool:
+        """Checks if this node has the given variation."""
+        return move in self
 
-    def promote_to_main(self, move: chess.Move) -> None:
+    def promote_to_main(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Promotes the given *move* to the main variation."""
         variation = self[move]
         self.variations.remove(variation)
         self.variations.insert(0, variation)
 
-    def promote(self, move: chess.Move) -> None:
+    def promote(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Moves a variation one up in the list of variations."""
         variation = self[move]
         i = self.variations.index(variation)
         if i > 0:
             self.variations[i - 1], self.variations[i] = self.variations[i], self.variations[i - 1]
 
-    def demote(self, move: chess.Move) -> None:
+    def demote(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Moves a variation one down in the list of variations."""
         variation = self[move]
         i = self.variations.index(variation)
         if i < len(self.variations) - 1:
             self.variations[i + 1], self.variations[i] = self.variations[i], self.variations[i + 1]
 
-    def remove_variation(self, move: chess.Move) -> None:
+    def remove_variation(self, move: Union[int, chess.Move, GameNode]) -> None:
         """Removes a variation."""
         self.variations.remove(self.variation(move))
 
-    def add_variation(self, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> "GameNode":
+    def add_variation(self, move: chess.Move, *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> GameNode:
         """Creates a child node with the given attributes."""
         node = type(self).dangling_node()
         node.move = move
@@ -346,7 +357,7 @@ class GameNode:
         self.variations.append(node)
         return node
 
-    def add_main_variation(self, move: chess.Move, *, comment: str = "") -> "GameNode":
+    def add_main_variation(self, move: chess.Move, *, comment: str = "") -> GameNode:
         """
         Creates a child node with the given attributes and promotes it to the
         main variation.
@@ -356,15 +367,15 @@ class GameNode:
         self.variations.insert(0, node)
         return node
 
-    def mainline(self) -> "Mainline[GameNode]":
+    def mainline(self) -> Mainline[GameNode]:
         """Returns an iterator over the mainline starting after this node."""
         return Mainline(self, lambda node: node)
 
-    def mainline_moves(self) -> "Mainline[chess.Move]":
+    def mainline_moves(self) -> Mainline[chess.Move]:
         """Returns an iterator over the main moves after this node."""
         return Mainline(self, lambda node: node._move())
 
-    def add_line(self, moves: Iterable[chess.Move], *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> "GameNode":
+    def add_line(self, moves: Iterable[chess.Move], *, comment: str = "", starting_comment: str = "", nags: Iterable[int] = []) -> GameNode:
         """
         Creates a sequence of child nodes for the given list of moves.
         Adds *comment* and *nags* to the last node of the line and returns it.
@@ -416,7 +427,7 @@ class GameNode:
         self.comment, found = EVAL_REGEX.subn(eval, self.comment, count=1)
 
         if not found and eval:
-            if not self.comment.endswith(" "):
+            if self.comment and not self.comment.endswith(" "):
                 self.comment += " "
             self.comment += eval
 
@@ -430,7 +441,7 @@ class GameNode:
         arrows = []
         for match in ARROWS_REGEX.finditer(self.comment):
             for group in match.group(1).split(","):
-                color = "G"
+                color = "green"
                 if group[0] == "R":
                     color = "red"
                 elif group[0] == "Y":
@@ -448,7 +459,7 @@ class GameNode:
 
         return arrows
 
-    def set_arrows(self, arrows: Iterable[Union[chess.svg.Arrow, Tuple[chess.Square, chess.Square]]]) -> None:
+    def set_arrows(self, arrows: Iterable[Union[chess.svg.Arrow, Tuple[Square, Square]]]) -> None:
         """
         Replaces all valid ``[%csl ...]`` and ``[%cal ...]`` annotations in
         the comment of this node or adds new ones.
@@ -484,7 +495,7 @@ class GameNode:
             prefix += f"[%cal {','.join(cal)}]"
 
         if prefix:
-            self.comment = prefix + " " + self.comment
+            self.comment = prefix + " " + self.comment if self.comment else prefix
 
     def clock(self) -> Optional[float]:
         """
@@ -506,6 +517,7 @@ class GameNode:
         """
         clk = ""
         if seconds is not None:
+            seconds = max(0, round(seconds))
             hours = seconds // 3600
             minutes = seconds % 3600 // 60
             seconds = seconds % 3600 % 60
@@ -514,11 +526,11 @@ class GameNode:
         self.comment, found = CLOCK_REGEX.subn(clk, self.comment, count=1)
 
         if not found and clk:
-            if not self.comment.endswith(" "):
+            if self.comment and not self.comment.endswith(" "):
                 self.comment += " "
             self.comment += clk
 
-    def _accept_node(self, parent_board: chess.Board, visitor: "BaseVisitor[ResultT]") -> None:
+    def _accept_node(self, parent_board: chess.Board, visitor: BaseVisitor[ResultT]) -> None:
         assert self.move is not None, "cannot visit dangling GameNode"
 
         if self.starting_comment:
@@ -536,7 +548,7 @@ class GameNode:
         if self.comment:
             visitor.visit_comment(self.comment)
 
-    def _accept(self, parent_board: chess.Board, visitor: "BaseVisitor[ResultT]", *, sidelines: bool = True) -> None:
+    def _accept(self, parent_board: chess.Board, visitor: BaseVisitor[ResultT], *, sidelines: bool = True) -> None:
         stack = [_AcceptFrame(self, sidelines=sidelines)]
 
         while stack:
@@ -570,7 +582,7 @@ class GameNode:
             else:
                 stack.pop()
 
-    def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         """
         Traverses game nodes in PGN order using the given *visitor*. Starts with
         the move leading to this node. Returns the *visitor* result.
@@ -579,7 +591,7 @@ class GameNode:
         self._accept(self.parent.board(), visitor, sidelines=False)
         return visitor.result()
 
-    def accept_subgame(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept_subgame(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         """
         Traverses headers and game nodes in PGN order, as if the game was
         starting after this node. Returns the *visitor* result.
@@ -675,7 +687,7 @@ class Game(GameNode):
         else:
             self.headers.pop("Variant", None)
 
-    def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         """
         Traverses the game in PGN order using the given *visitor*. Returns
         the *visitor* result.
@@ -719,7 +731,7 @@ class Game(GameNode):
         return cls(headers={})
 
     @classmethod
-    def builder(cls: Type[GameT]) -> "GameBuilder[GameT]":
+    def builder(cls: Type[GameT]) -> GameBuilder[GameT]:
         return GameBuilder(Game=cls)
 
     def __repr__(self) -> str:
@@ -825,7 +837,7 @@ class Headers(MutableMapping[str, str]):
             ", ".join("{}={!r}".format(key, value) for key, value in self.items()))
 
     @classmethod
-    def builder(cls: Type[HeadersT]) -> "HeadersBuilder[HeadersT]":
+    def builder(cls: Type[HeadersT]) -> HeadersBuilder[HeadersT]:
         return HeadersBuilder(Headers=cls)
 
 
@@ -845,10 +857,10 @@ class Mainline(Generic[MainlineMapT]):
             node = node.variations[0]
             yield self.f(node)
 
-    def __reversed__(self) -> "ReverseMainline[MainlineMapT]":
+    def __reversed__(self) -> ReverseMainline[MainlineMapT]:
         return ReverseMainline(self.start, self.f)
 
-    def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
+    def accept(self, visitor: BaseVisitor[ResultT]) -> ResultT:
         node = self.start
         board = self.start.board()
         while node.variations:
@@ -997,10 +1009,10 @@ class GameBuilder(BaseVisitor[GameT]):
     """
 
     @typing.overload
-    def __init__(self: "GameBuilder[Game]") -> None: ...
+    def __init__(self: GameBuilder[Game]) -> None: ...
     @typing.overload
-    def __init__(self: "GameBuilder[GameT]", *, Game: Type[GameT]) -> None: ...
-    def __init__(self, *, Game = Game) -> None:
+    def __init__(self: GameBuilder[GameT], *, Game: Type[GameT]) -> None: ...
+    def __init__(self, *, Game: Any = Game) -> None:
         self.Game = Game
 
     def begin_game(self) -> None:
@@ -1073,7 +1085,7 @@ class GameBuilder(BaseVisitor[GameT]):
         >>> import chess.pgn
         >>>
         >>> class MyGameBuilder(chess.pgn.GameBuilder):
-        >>>     def handle_error(self, error):
+        >>>     def handle_error(self, error: Exception) -> None:
         >>>         pass  # Ignore error
         >>>
         >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn")
@@ -1094,10 +1106,10 @@ class HeadersBuilder(BaseVisitor[HeadersT]):
     """Collects headers into a dictionary."""
 
     @typing.overload
-    def __init__(self: "HeadersBuilder[Headers]") -> None: ...
+    def __init__(self: HeadersBuilder[Headers]) -> None: ...
     @typing.overload
-    def __init__(self: "HeadersBuilder[HeadersT]", *, Headers: Type[Headers]) -> None: ...
-    def __init__(self, *, Headers = Headers) -> None:
+    def __init__(self: HeadersBuilder[HeadersT], *, Headers: Type[Headers]) -> None: ...
+    def __init__(self, *, Headers: Any = Headers) -> None:
         self.Headers = Headers
 
     def begin_headers(self) -> HeadersT:
@@ -1290,7 +1302,7 @@ class FileExporter(StringExporterMixin, BaseVisitor[int]):
         super().__init__(columns=columns, headers=headers, comments=comments, variations=variations)
         self.handle = handle
 
-    def begin_game(self):
+    def begin_game(self) -> None:
         self.written: int = 0
         super().begin_game()
 
@@ -1319,7 +1331,7 @@ class FileExporter(StringExporterMixin, BaseVisitor[int]):
 def read_game(handle: TextIO) -> Optional[Game]: ...
 @typing.overload
 def read_game(handle: TextIO, *, Visitor: Callable[[], BaseVisitor[ResultT]]) -> Optional[ResultT]: ...
-def read_game(handle: TextIO, *, Visitor = GameBuilder):
+def read_game(handle: TextIO, *, Visitor: Any = GameBuilder) -> Any:
     """
     Reads a game from a file opened in text mode.
 
@@ -1343,10 +1355,10 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
 
     By using text mode, the parser does not need to handle encodings. It is the
     caller's responsibility to open the file with the correct encoding.
-    PGN files are usually ASCII or UTF-8 encoded. So, the following should
-    cover most relevant cases (ASCII, UTF-8, UTF-8 with BOM).
+    PGN files are usually ASCII or UTF-8 encoded, sometimes with BOM (which
+    this parser automatically ignores).
 
-    >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn", encoding="utf-8-sig")
+    >>> pgn = open("data/pgn/kasparov-deep-blue-1997.pgn", encoding="utf-8")
 
     Use :class:`~io.StringIO` to parse games from a string.
 
@@ -1365,7 +1377,7 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
     The parser is relatively forgiving when it comes to errors. It skips over
     tokens it can not parse. By default, any exceptions are logged and
     collected in :data:`Game.errors <chess.pgn.Game.errors>`. This behavior can
-    be :func:`overriden <chess.pgn.GameBuilder.handle_error>`.
+    be :func:`overridden <chess.pgn.GameBuilder.handle_error>`.
 
     Returns the parsed game or ``None`` if the end of file is reached.
     """
