@@ -14,6 +14,8 @@
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
+from __future__ import annotations
+
 import chess
 import copy
 import itertools
@@ -375,18 +377,14 @@ class RacingKingsBoard(chess.Board):
         if not self.kings & chess.BB_RANK_8:
             return False
 
-        if self.turn == chess.WHITE or self.kings & self.occupied_co[chess.BLACK] & chess.BB_RANK_8:
-            return True
-
         black_kings = self.kings & self.occupied_co[chess.BLACK]
-        if not black_kings:
+        if self.turn == chess.WHITE or black_kings & chess.BB_RANK_8 or not black_kings:
             return True
 
-        black_king = chess.msb(black_kings)
-
         # White has reached the backrank. The game is over if black can not
         # also reach the backrank on the next move. Check if there are any
         # safe squares for the king.
+        black_king = chess.msb(black_kings)
         targets = chess.BB_KING_ATTACKS[black_king] & chess.BB_RANK_8 & ~self.occupied_co[chess.BLACK]
         return all(self.attackers_mask(chess.WHITE, target) for target in chess.scan_forward(targets))
 
@@ -398,7 +396,11 @@ class RacingKingsBoard(chess.Board):
         return self.is_variant_end() and not self.kings & self.occupied_co[self.turn] & chess.BB_RANK_8
 
     def is_variant_win(self) -> bool:
-        return self.is_variant_end() and bool(self.kings & self.occupied_co[self.turn] & chess.BB_RANK_8)
+        in_goal = self.kings & chess.BB_RANK_8
+        return (
+            self.is_variant_end() and
+            bool(in_goal & self.occupied_co[self.turn]) and
+            not in_goal & self.occupied_co[not self.turn])
 
     def has_insufficient_material(self, color: chess.Color) -> bool:
         return False
@@ -476,13 +478,13 @@ class HordeBoard(chess.Board):
 
 ThreeCheckBoardT = TypeVar("ThreeCheckBoardT", bound="ThreeCheckBoard")
 
-class _ThreeCheckBoardState(Generic[ThreeCheckBoardT], chess._BoardState["ThreeCheckBoardT"]):
-    def __init__(self, board: "ThreeCheckBoardT") -> None:
+class _ThreeCheckBoardState(Generic[ThreeCheckBoardT], chess._BoardState[ThreeCheckBoardT]):
+    def __init__(self, board: ThreeCheckBoardT) -> None:
         super().__init__(board)
         self.remaining_checks_w = board.remaining_checks[chess.WHITE]
         self.remaining_checks_b = board.remaining_checks[chess.BLACK]
 
-    def restore(self, board: "ThreeCheckBoardT") -> None:
+    def restore(self, board: ThreeCheckBoardT) -> None:
         super().restore(board)
         board.remaining_checks[chess.WHITE] = self.remaining_checks_w
         board.remaining_checks[chess.BLACK] = self.remaining_checks_b
@@ -606,13 +608,13 @@ class ThreeCheckBoard(chess.Board):
 
 CrazyhouseBoardT = TypeVar("CrazyhouseBoardT", bound="CrazyhouseBoard")
 
-class _CrazyhouseBoardState(Generic[CrazyhouseBoardT], chess._BoardState["CrazyhouseBoardT"]):
-    def __init__(self, board: "CrazyhouseBoardT") -> None:
+class _CrazyhouseBoardState(Generic[CrazyhouseBoardT], chess._BoardState[CrazyhouseBoardT]):
+    def __init__(self, board: CrazyhouseBoardT) -> None:
         super().__init__(board)
         self.pockets_w = board.pockets[chess.WHITE].copy()
         self.pockets_b = board.pockets[chess.BLACK].copy()
 
-    def restore(self, board: "CrazyhouseBoardT") -> None:
+    def restore(self, board: CrazyhouseBoardT) -> None:
         super().restore(board)
         board.pockets[chess.WHITE] = self.pockets_w.copy()
         board.pockets[chess.BLACK] = self.pockets_b.copy()
@@ -698,10 +700,8 @@ class CrazyhouseBoard(chess.Board):
         else:
             self.pockets[self.turn].add(piece_type)
 
-    def can_claim_fifty_moves(self) -> bool:
-        return False
-
-    def is_seventyfive_moves(self) -> bool:
+    def _is_halfmoves(self, n: int) -> bool:
+        # No draw by 50-move rule or 75-move rule.
         return False
 
     def is_irreversible(self, move: chess.Move) -> bool:
